package com.b3.rbl.compensation.service;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.net.URI;
import java.nio.file.Path;
import java.util.List;
import java.util.Map;
import java.util.UUID;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Date;

import org.apache.http.HttpEntity;
import org.apache.http.client.methods.CloseableHttpResponse;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.impl.client.CloseableHttpClient;
import org.apache.http.impl.client.HttpClients;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Service;
import org.springframework.util.StringUtils;

import com.b3.rbl.compensation.db.entity.Variable;
import com.b3.rbl.compensation.db.entity.RblAgentReport;
import com.b3.rbl.compensation.db.repository.RblAgenReportRepository;
import com.b3.rbl.compensation.db.repository.VariableRepository;
import com.b3.rbl.compensation.model.EmailSenderModel;
import com.b3.rbl.compensation.model.FileAttachmentModel;
import com.b3.rbl.compensation.util.CompileService;
import com.b3.rbl.onedrive.client.Client;
import com.b3.rbl.onedrive.container.RblDriveItem;
import com.b3.rbl.onedrive.container.RblDriveItemSet;
import com.b3.rbl.onedrive.container.items.pointer.PathPointer;
import com.b3.rbl.onedrive.exceptions.ErrorResponseException;
import com.fasterxml.jackson.core.type.TypeReference;
import com.fasterxml.jackson.databind.ObjectMapper;

import lombok.extern.log4j.Log4j2;

@Log4j2
@Service
public class OnedriveService {
    private VariableRepository variableRepository;
    private RblAgenReportRepository rblAgenReportRepository;
    private Client client;
    private EmailService emailService;
    private String reportOnedrivePathStatementPrefix;
    private static final ObjectMapper mapper = new ObjectMapper();

    public OnedriveService(@Value("${report.onedrive.path.statement.prefix}") String reportOnedrivePathPrefix,
            EmailService emailService, VariableRepository variableRepository,
            RblAgenReportRepository rblAgenReportRepository, Client client) {
        this.reportOnedrivePathStatementPrefix = reportOnedrivePathPrefix;
        this.variableRepository = variableRepository;
        this.rblAgenReportRepository = rblAgenReportRepository;
        this.client = client;
        this.emailService = emailService;
    }

    public void saveFile(String parentPath, String fileName, Path filePath) throws IOException, ErrorResponseException {
        PathPointer pathPointer = new PathPointer(parentPath);
        client.simpleUploadFile(pathPointer, fileName, filePath);

    }

    public List<RblAgentReport> getFileFromPath(String year, String month, String round) throws Exception {
        String[] paths = { "01_Statement", "02_Hold" };
        List<RblAgentReport> entities = new ArrayList<>();
        Integer pyear = Integer.parseInt(year) + 543;
        UUID guid = UUID.randomUUID();
        Date now = new Date();
        for (String p : paths) {
            String path = String.format("%s/%s/%s/%s/%s", reportOnedrivePathStatementPrefix, pyear, month, round, p);
            RblDriveItemSet itemset = this.client.listFileItem(path, Collections.singletonMap("top", "1000"));
            for (RblDriveItem item : itemset.getValue()) {
                RblAgentReport entity = deserialize(p, pyear, month, round, item);
                entity.setGuid(guid);
                entity.setCreateDate(now);
                entities.add(entity);
            }
        }

        rblAgenReportRepository.clearReport(pyear.toString(), month, round);
        rblAgenReportRepository.saveAll(entities);

        return entities;
    }

    private RblAgentReport deserialize(String reportName, Integer year, String month, String round, RblDriveItem item) {

        RblAgentReport entity = new RblAgentReport();
        try {

            String[] regx = item.getName().split("_");
            String line = regx[0];
            String agentCode = regx[1];
            String[] lineInfo = line.split("-");
            // String[] pathElem = item.getWebUrl().split("/");
            // int size = pathElem.length;
            // String _reportname = pathElem[--size];
            // String _period = pathElem[--size];
            // String _period_month = pathElem[--size];
            // String _period_year = pathElem[--size];

            entity.setReportName(reportName);
            entity.setPeriod(round);
            entity.setPeriodMonth(month);
            entity.setPeriodYear(year.toString());
            entity.setFilename(item.getName());
            entity.setMimeType(item.getFile().getMimeType());
            entity.setAgentCode(agentCode);
            entity.setSize(item.getSize());
            // entity.setLineName(lineInfo[0]);
            entity.setLineCode(lineInfo[0]);
            entity.setGroupCode(lineInfo[1]);
            entity.setUnitCode(lineInfo[2]);
            entity.setDriveId(item.getParentReference().getDriveId());
            entity.setItemId(item.getId());
            entity.setdownloadUrl(item.getDownloadUrl());
            log.info("AgentCode[{}] Period: {} {} {}  ", agentCode, year, month, round);
        } catch (NullPointerException | ArrayIndexOutOfBoundsException ex) {
            log.error("Error on record {}", item.getWebUrl());
        }

        return entity;
    }

    public void sendemail(List<RblAgentReport> agentReport) throws Exception {
        Variable vsubject = variableRepository.findByKey(STATMENT_EMAIL_SUBJECT);
        Variable vtemplate = variableRepository.findByKey(STATMENT_EMAIL_BODY);

        for (RblAgentReport data : agentReport) {
            EmailSenderModel email = buildEmailTemplate(data, vsubject.getValue(), vtemplate.getValue());
            emailService.sendEmail(email);
        }
    }

    public byte[] getFileFromUrl(String downloadUrl) throws Exception {
        byte[] data = null;
        try (CloseableHttpClient httpClient = HttpClients.createDefault()) {
            log.info("downloadUrl {} ", downloadUrl);
            HttpGet httpget = new HttpGet(URI.create(downloadUrl));
            httpget.addHeader("Authorization", this.client.getFullToken());
            try (CloseableHttpResponse response = httpClient.execute(httpget)) {

                HttpEntity entity = response.getEntity();
                log.info("response.getStatusLine().getStatusCode {} ", response.getStatusLine().getStatusCode());
                if (response.getStatusLine().getStatusCode() > 200) {
                    throw new Exception("Please verify file datain " + downloadUrl);
                }
                if (entity != null) {
                    ByteArrayOutputStream baos = new ByteArrayOutputStream();
                    entity.writeTo(baos);
                    data = baos.toByteArray();
                }

            }
        }
        return data;
    }

    public EmailSenderModel buildEmailTemplate(RblAgentReport data, String subject, String body) throws Exception {

        FileAttachmentModel attachment = null;
        if (StringUtils.hasLength(data.getdownloadUrl())) {

            data.getAgentCode();
            byte[] input = getFileFromUrl(data.getdownloadUrl());
            attachment = new FileAttachmentModel(input, data.getFilename());

        }
        Map<String, Object> variable = OnedriveService.mapper.convertValue(data,
                new TypeReference<Map<String, Object>>() {
                });

        body = CompileService.instanct().dataJson(body, variable);
        subject = CompileService.instanct().dataJson(subject, variable);
        return new EmailSenderModel(new String[] {}, new String[] {}, body, subject, Arrays.asList(attachment));
    }

    private static final String STATMENT_EMAIL_SUBJECT = "STATMENT_EMAIL_SUBJECT";
    private static final String STATMENT_EMAIL_BODY = "STATMENT_EMAIL_BODY";

}
