package com.b3.rbl.onedrive.container.items;

import com.b3.rbl.onedrive.client.Client;
import com.b3.rbl.onedrive.container.facet.RemoteItemFacet;
import com.b3.rbl.onedrive.container.facet.SearchResultFacet;
import com.b3.rbl.onedrive.container.facet.SharePointIdsFacet;
import com.b3.rbl.onedrive.container.facet.SharedFacet;
import com.b3.rbl.onedrive.container.items.pointer.Operator;
import com.b3.rbl.onedrive.exceptions.ErrorResponseException;
import com.b3.rbl.onedrive.network.async.DriveItemFuture;
import com.b3.rbl.onedrive.network.async.ResponseFuture;
import com.b3.rbl.onedrive.network.async.UploadFuture;
import io.netty.handler.codec.http.HttpMethod;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.io.IOException;
import java.net.URI;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import static java.net.HttpURLConnection.HTTP_OK;


public class RemoteFolderItem extends AbstractRemoteItem implements FolderItem {
	protected List<FolderItem> folderChildren;
	protected List<FileItem> fileChildren;
	protected List<DriveItem> allChildren;

	RemoteFolderItem(@NotNull String id, @NotNull String createdDateTime, @Nullable String description,
					 @NotNull String eTag, @NotNull String lastModifiedDateTime, @NotNull String name,
					 @NotNull URI webUrl, @NotNull Client client, @NotNull String cTag, @Nullable String deleted,
					 @NotNull ItemReference parentReference, @Nullable SearchResultFacet searchResult,
					 @Nullable SharedFacet shared, @Nullable SharePointIdsFacet sharePointIds, URI webDavUrl,
					 @NotNull RemoteItemFacet remoteItem) {
		super(id, createdDateTime, description, eTag, lastModifiedDateTime, name, webUrl, client, cTag, deleted,
				parentReference, searchResult, shared, sharePointIds, webDavUrl, remoteItem);
	}

	@Override
	public @NotNull FolderItem createFolder(@NotNull String name) throws ErrorResponseException {
		return client.createFolder(remotePointer, name);
	}

	@Override public @NotNull UploadFuture uploadFile(@NotNull Path filePath) {
		return client.uploadFile(this.remoteItem.getId(), filePath);
	}

	@Override
	public @NotNull FileItem simpleUploadFile(@NotNull Path filePath) throws IOException, ErrorResponseException {
		return client.simpleUploadFile(this.remoteItem.getId(), filePath);
	}

	@Override public @NotNull DriveItemFuture simpleUploadFileAsync(@NotNull Path filePath) {
		return client.simpleUploadFileAsync(this.remoteItem.getId(), filePath);
	}

	@Override public boolean isDeleted() {return deleted != null;}

	@Override public @Nullable String deletedState() {return deleted;}

	@Override public boolean isRoot() {return false;}

	@Override public boolean isChildrenFetched() {
		assert remoteItem.getFolder() != null : "remoteItem.getFolder() is null";
		return remoteItem.getFolder().getChildCount() == 0
				|| allChildren != null && folderChildren != null && fileChildren != null;
	}

	protected void addChildren(@NotNull DriveItem[] array) {
		for (DriveItem item : array) {
			if (item instanceof FolderItem) {
				folderChildren.add((FolderItem) item);
			}
			else if (item instanceof FileItem) {
				fileChildren.add((FileItem) item);
			}
			else {
				// if child is neither FolderItem nor FileItem nor PackageItem.
				assert item instanceof PackageItem : "Wrong item type";
			}
			allChildren.add(item);
		}
	}

	@Override public void fetchChildren() throws ErrorResponseException {
		allChildren = new ArrayList<>();
		folderChildren = new ArrayList<>();
		fileChildren = new ArrayList<>();

		ResponseFuture responseFuture = client.requestTool()
				.doAsync(HttpMethod.GET, remotePointer.resolveOperator(Operator.CHILDREN))
				.syncUninterruptibly();

		addChildren(client.requestTool()
				.parseDriveItemRecursiveAndHandle(responseFuture.response(), responseFuture.getNow(), HTTP_OK));
	}

	@Override public boolean isSpecial() {return false;}

	@Override public long childCount() {
		assert remoteItem.getFolder() != null : "remoteItem.getFolder() is null";
		return remoteItem.getFolder().getChildCount();
	}

	@Override public @NotNull DriveItem[] allChildren() throws ErrorResponseException {
		if (!isChildrenFetched()) fetchChildren();
		return allChildren.toArray(new DriveItem[0]);
	}

	@Override public @NotNull FolderItem[] folderChildren() throws ErrorResponseException {
		if (!isChildrenFetched()) fetchChildren();
		return folderChildren.toArray(new FolderItem[0]);
	}

	@Override public @NotNull FileItem[] fileChildren() throws ErrorResponseException {
		if (!isChildrenFetched()) fetchChildren();
		return fileChildren.toArray(new FileItem[0]);
	}

	@Override public @NotNull Iterator<DriveItem> iterator() {
		try {
			if (!isChildrenFetched()) fetchChildren();
			return allChildren.iterator();
		}
		catch (ErrorResponseException e) {
			// FIXME: custom exception
			throw new RuntimeException(e);
		}
	}
}
